@@ -8,7 +8,7 @@
  *
  *    In some countries OX, OX Open-Xchange, open xchange and OXtender
  *    as well as the corresponding Logos OX Open-Xchange and OX are registered
- *    trademarks of the Open-Xchange, Inc. group of companies.
+ *    trademarks of the OX Software GmbH group of companies.
  *    The use of the Logos is not covered by the GNU General Public License.
  *    Instead, you are allowed to use these Logos according to the terms and
  *    conditions of the Creative Commons License, Version 2.5, Attribution,
@@ -28,7 +28,7 @@
  *    http://www.open-xchange.com/EN/developer/. The contributing author shall be
  *    given Attribution for the derivative code and a license granting use.
  *
- *     Copyright (C) 2004-2014 Open-Xchange, Inc.
+ *     Copyright (C) 2016-2020 OX Software GmbH
  *     Mail: info@open-xchange.com
  *
  *
@@ -53,7 +53,9 @@ import static com.openexchange.mail.MailExceptionCode.getSize;
 import static com.openexchange.mail.MailServletInterface.mailInterfaceMonitor;
 import static com.openexchange.mail.mime.utils.MimeMessageUtility.parseAddressList;
 import static com.openexchange.mail.text.TextProcessing.performLineFolding;
+
 import java.io.IOException;
+import java.io.InputStream;
 import java.io.UnsupportedEncodingException;
 import java.net.InetAddress;
 import java.net.InetSocketAddress;
@@ -70,6 +72,7 @@ import java.util.Properties;
 import java.util.Queue;
 import java.util.concurrent.ConcurrentLinkedQueue;
 import java.util.regex.Matcher;
+
 import javax.activation.DataHandler;
 import javax.mail.Address;
 import javax.mail.Message.RecipientType;
@@ -88,6 +91,7 @@ import javax.mail.internet.MimeMessage;
 import javax.mail.internet.MimeMultipart;
 import javax.mail.internet.idn.IDNA;
 import javax.security.auth.Subject;
+
 import com.openexchange.config.ConfigurationService;
 import com.openexchange.config.Filter;
 import com.openexchange.config.cascade.ConfigProperty;
@@ -103,6 +107,7 @@ import com.openexchange.i18n.tools.StringHelper;
 import com.openexchange.java.Charsets;
 import com.openexchange.java.Java7ConcurrentLinkedQueue;
 import com.openexchange.java.Strings;
+import com.openexchange.java.UnsynchronizedByteArrayInputStream;
 import com.openexchange.java.util.MsisdnCheck;
 import com.openexchange.log.LogProperties;
 import com.openexchange.mail.MailExceptionCode;
@@ -129,6 +134,8 @@ import com.openexchange.mail.transport.MtaStatusInfo;
 import com.openexchange.mail.transport.config.ITransportProperties;
 import com.openexchange.mail.transport.config.TransportConfig;
 import com.openexchange.mail.transport.config.TransportProperties;
+import com.openexchange.mail.transport.listener.Reply;
+import com.openexchange.mail.transport.listener.Result;
 import com.openexchange.mail.usersetting.UserSettingMail;
 import com.openexchange.mail.usersetting.UserSettingMailStorage;
 import com.openexchange.mail.utils.MessageUtility;
@@ -382,6 +389,12 @@ public final class SMTPTransport extends MailTransport implements MimeSupport {
                     if (smtpProperties.getSmtpConnectionTimeout() > 0) {
                         smtpProps.put("mail.smtp.connectiontimeout", Integer.toString(smtpProperties.getSmtpConnectionTimeout()));
                     }
+                    /*
+                     * Send partial or abort?
+                     */
+                    if (smtpProperties.isSendPartial()) {
+                        smtpProps.put("mail.smtp.sendpartial", "true");
+                    }
                     /*
                      * Check if a secure SMTP connection should be established
                      */
@@ -480,7 +493,7 @@ public final class SMTPTransport extends MailTransport implements MimeSupport {
                         Provider.Type.TRANSPORT,
                         "smtp",
                         JavaSMTPTransport.class.getName(),
-                        "Open-Xchange, Inc.",
+                        "OX Software GmbH",
                         "7.2.2"));
                 }
             }
@@ -520,7 +533,7 @@ public final class SMTPTransport extends MailTransport implements MimeSupport {
     private static final String MULTI_SUBTYPE_REPORT = "report; report-type=disposition-notification";
 
     @Override
-    public void sendReceiptAck(final MailMessage srcMail, final String fromAddr) throws OXException {
+    public void sendReceiptAck(MailMessage srcMail, String fromAddr) throws OXException {
         if (null == srcMail) {
             return;
         }
@@ -534,16 +547,19 @@ public final class SMTPTransport extends MailTransport implements MimeSupport {
             // Set the ClassLoader to the javax.mail bundle loader.
             // Thread.currentThread().setContextClassLoader(MailMessage.class.getClassLoader());
             // Go ahead...
-            final InternetAddress dispNotification = srcMail.getDispositionNotification();
+            InternetAddress dispNotification = srcMail.getDispositionNotification();
             if (dispNotification == null) {
-                throw SMTPExceptionCode.MISSING_NOTIFICATION_HEADER.create(MessageHeaders.HDR_DISP_TO, Long.valueOf(srcMail.getMailId()));
+                dispNotification = srcMail.getFrom()[0];
+                if (null == dispNotification) {
+                    throw SMTPExceptionCode.MISSING_NOTIFICATION_HEADER.create(MessageHeaders.HDR_DISP_TO, Long.valueOf(srcMail.getMailId()));
+                }
             }
-            final SMTPMessage smtpMessage = new SMTPMessage(getSMTPSession());
-            final String userMail = UserStorage.getInstance().getUser(session.getUserId(), ctx).getMail();
+            SMTPMessage smtpMessage = new SMTPMessage(getSMTPSession());
+            String userMail = UserStorage.getInstance().getUser(session.getUserId(), ctx).getMail();
             /*
              * Set from
              */
-            final String from;
+            String from;
             if (fromAddr == null) {
                 if ((usm.getSendAddr() == null) && (userMail == null)) {
                     throw SMTPExceptionCode.NO_SEND_ADDRESS_FOUND.create();
@@ -556,7 +572,7 @@ public final class SMTPTransport extends MailTransport implements MimeSupport {
             /*
              * Set to
              */
-            final Address[] recipients = new Address[] { dispNotification };
+            Address[] recipients = new Address[] { dispNotification };
             processAddressHeader(smtpMessage);
             checkRecipients(recipients);
             smtpMessage.addRecipients(RecipientType.TO, recipients);
@@ -568,14 +584,14 @@ public final class SMTPTransport extends MailTransport implements MimeSupport {
             /*
              * Subject
              */
-            final Locale locale = UserStorage.getInstance().getUser(session.getUserId(), ctx).getLocale();
-            final StringHelper strHelper = StringHelper.valueOf(locale);
+            Locale locale = UserStorage.getInstance().getUser(session.getUserId(), ctx).getLocale();
+            StringHelper strHelper = StringHelper.valueOf(locale);
             smtpMessage.setSubject(strHelper.getString(MailStrings.ACK_SUBJECT));
             /*
              * Sent date in UTC time
              */
             {
-                final MailDateFormat mdf = MimeMessageUtility.getMailDateFormat(session);
+                MailDateFormat mdf = MimeMessageUtility.getMailDateFormat(session);
                 synchronized (mdf) {
                     smtpMessage.setHeader("Date", mdf.format(new Date()));
                 }
@@ -588,19 +604,23 @@ public final class SMTPTransport extends MailTransport implements MimeSupport {
             /*
              * Compose body
              */
-            final String defaultMimeCS = MailProperties.getInstance().getDefaultMimeCharset();
-            final ContentType ct = new ContentType(CT_TEXT_PLAIN.replaceFirst("#CS#", defaultMimeCS));
-            final Multipart mixedMultipart = new MimeMultipart(MULTI_SUBTYPE_REPORT);
+            String defaultMimeCS = MailProperties.getInstance().getDefaultMimeCharset();
+            ContentType ct = new ContentType(CT_TEXT_PLAIN.replaceFirst("#CS#", defaultMimeCS));
+            Multipart mixedMultipart = new MimeMultipart(MULTI_SUBTYPE_REPORT);
             /*
              * Define text content
              */
-            final Date sentDate = srcMail.getSentDate();
+            Date sentDate = srcMail.getSentDate();
             {
-                final MimeBodyPart text = new MimeBodyPart();
-                final String txt = performLineFolding(
-                    strHelper.getString(MailStrings.ACK_NOTIFICATION_TEXT)
-                    .replaceFirst("#DATE#", sentDate == null ? "" : quoteReplacement(DateFormat.getDateInstance(DateFormat.LONG, locale).format(sentDate)))
-                    .replaceFirst("#RECIPIENT#", quoteReplacement(from)).replaceFirst("#SUBJECT#", quoteReplacement(srcMail.getSubject())), usm.getAutoLinebreak());
+                MimeBodyPart text = new MimeBodyPart();
+                String txt =
+                    performLineFolding(
+                        strHelper.getString(MailStrings.ACK_NOTIFICATION_TEXT).replaceFirst(
+                            "#DATE#",
+                            sentDate == null ? "" : quoteReplacement(DateFormat.getDateInstance(DateFormat.LONG, locale).format(sentDate))).replaceFirst(
+                            "#RECIPIENT#",
+                            quoteReplacement(from)).replaceFirst("#SUBJECT#", quoteReplacement(srcMail.getSubject())),
+                        usm.getAutoLinebreak());
                 MessageUtility.setText(txt, defaultMimeCS, text);
                 // text.setText(txt,defaultMimeCS);
                 text.setHeader(MessageHeaders.HDR_MIME_VERSION, "1.0");
@@ -612,11 +632,9 @@ public final class SMTPTransport extends MailTransport implements MimeSupport {
              */
             ct.setContentType(CT_READ_ACK);
             {
-                final MimeBodyPart ack = new MimeBodyPart();
-                final String msgId = srcMail.getFirstHeader(MessageHeaders.HDR_MESSAGE_ID);
-                final String txt = strHelper.getString(ACK_TEXT).replaceFirst("#FROM#", quoteReplacement(from)).replaceFirst(
-                    "#MSG ID#",
-                    quoteReplacement(msgId));
+                MimeBodyPart ack = new MimeBodyPart();
+                String msgId = srcMail.getFirstHeader(MessageHeaders.HDR_MESSAGE_ID);
+                String txt = strHelper.getString(ACK_TEXT).replaceFirst("#FROM#", quoteReplacement(from)).replaceFirst("#MSG ID#", quoteReplacement(msgId));
                 MessageUtility.setText(txt, defaultMimeCS, ack);
                 // ack.setText(txt,defaultMimeCS);
                 ack.setHeader(MessageHeaders.HDR_MIME_VERSION, "1.0");
@@ -632,19 +650,19 @@ public final class SMTPTransport extends MailTransport implements MimeSupport {
             /*
              * Transport message
              */
-            final long start = System.currentTimeMillis();
-            final Transport transport = getSMTPSession().getTransport(SMTP);
+            long start = System.currentTimeMillis();
+            Transport transport = getSMTPSession().getTransport(SMTP);
             try {
                 connectTransport(transport, smtpConfig);
-                saveChangesSafe(smtpMessage);
+                saveChangesSafe(smtpMessage, true);
                 transport(smtpMessage, smtpMessage.getAllRecipients(), transport, smtpConfig);
                 mailInterfaceMonitor.addUseTime(System.currentTimeMillis() - start);
-            } catch (final javax.mail.AuthenticationFailedException e) {
+            } catch (javax.mail.AuthenticationFailedException e) {
                 throw MimeMailExceptionCode.TRANSPORT_INVALID_CREDENTIALS.create(e, smtpConfig.getServer(), e.getMessage());
             } finally {
                 transport.close();
             }
-        } catch (final MessagingException e) {
+        } catch (MessagingException e) {
             throw MimeMailException.handleMessagingException(e, smtpConfig, session);
         } finally {
             // Restore the ClassLoader
@@ -652,6 +670,58 @@ public final class SMTPTransport extends MailTransport implements MimeSupport {
         }
     }
 
+    @Override
+    public MailMessage sendRawMessage(byte[] asciiBytes, SendRawProperties properties) throws OXException {
+        final SMTPConfig smtpConfig = getTransportConfig0();
+        try {
+            InputStream rfc822IS;
+            if (properties.isSanitizeHeaders()) {
+                rfc822IS = MimeHeaderNameChecker.sanitizeHeaderNames(asciiBytes);
+            } else {
+                rfc822IS = new UnsynchronizedByteArrayInputStream(asciiBytes);
+            }
+            final SMTPMessage smtpMessage = new SMTPMessage(getSMTPSession(), rfc822IS);
+            InternetAddress sender = properties.getSender();
+            if (sender != null) {
+                smtpMessage.setEnvelopeFrom(sender.getAddress());
+            }
+            /*
+             * Check recipients
+             */
+            Address[] recipients = properties.getRecipients();
+            if (recipients == null) {
+                recipients = smtpMessage.getAllRecipients();
+            }
+            if (properties.isValidateAddressHeaders()) {
+                processAddressHeader(smtpMessage);
+            }
+            final boolean poisoned = checkRecipients(recipients);
+            if (poisoned) {
+                saveChangesSafe(smtpMessage, true);
+            } else {
+                try {
+                    final long start = System.currentTimeMillis();
+                    final Transport transport = getSMTPSession().getTransport(SMTP);
+                    try {
+                        connectTransport(transport, smtpConfig);
+                        saveChangesSafe(smtpMessage, true);
+                        transport(smtpMessage, recipients, transport, smtpConfig);
+                        mailInterfaceMonitor.addUseTime(System.currentTimeMillis() - start);
+                    } catch (final javax.mail.AuthenticationFailedException e) {
+                        throw MimeMailExceptionCode.TRANSPORT_INVALID_CREDENTIALS.create(e, smtpConfig.getServer(), e.getMessage());
+                    } finally {
+                        transport.close();
+                    }
+                } catch (final MessagingException e) {
+                    throw MimeMailException.handleMessagingException(e, smtpConfig, session);
+                }
+            }
+            return MimeMessageConverter.convertMessage(smtpMessage);
+        } catch (final MessagingException e) {
+            throw MimeMailException.handleMessagingException(e, smtpConfig, session);
+        }
+    }
+
     @Override
     public MailMessage sendRawMessage(final byte[] asciiBytes, final Address[] allRecipients) throws OXException {
         final SMTPConfig smtpConfig = getTransportConfig0();
@@ -673,14 +743,14 @@ public final class SMTPTransport extends MailTransport implements MimeSupport {
             processAddressHeader(smtpMessage);
             final boolean poisoned = checkRecipients(recipients);
             if (poisoned) {
-                saveChangesSafe(smtpMessage);
+                saveChangesSafe(smtpMessage, true);
             } else {
                 try {
                     final long start = System.currentTimeMillis();
                     final Transport transport = getSMTPSession().getTransport(SMTP);
                     try {
                         connectTransport(transport, smtpConfig);
-                        saveChangesSafe(smtpMessage);
+                        saveChangesSafe(smtpMessage, true);
                         transport(smtpMessage, recipients, transport, smtpConfig);
                         mailInterfaceMonitor.addUseTime(System.currentTimeMillis() - start);
                     } catch (final javax.mail.AuthenticationFailedException e) {
@@ -712,14 +782,14 @@ public final class SMTPTransport extends MailTransport implements MimeSupport {
             processAddressHeader(mimeMessage);
             final boolean poisoned = checkRecipients(recipients);
             if (poisoned) {
-                saveChangesSafe(mimeMessage);
+                saveChangesSafe(mimeMessage, true);
             } else {
                 try {
                     final long start = System.currentTimeMillis();
                     final Transport transport = getSMTPSession().getTransport(SMTP);
                     try {
                         connectTransport(transport, smtpConfig);
-                        saveChangesSafe(mimeMessage);
+                        saveChangesSafe(mimeMessage, true);
                         transport(mimeMessage, recipients, transport, smtpConfig);
                         mailInterfaceMonitor.addUseTime(System.currentTimeMillis() - start);
                     } catch (final javax.mail.AuthenticationFailedException e) {
@@ -816,14 +886,14 @@ public final class SMTPTransport extends MailTransport implements MimeSupport {
                 processAddressHeader(mimeMessage);
                 final boolean poisoned = checkRecipients(recipients);
                 if (poisoned) {
-                    saveChangesSafe(mimeMessage);
+                    saveChangesSafe(mimeMessage, true);
                 } else {
                     try {
                         final long start = System.currentTimeMillis();
                         final Transport transport = getSMTPSession().getTransport(SMTP);
                         try {
                             connectTransport(transport, smtpConfig);
-                            saveChangesSafe(mimeMessage);
+                            saveChangesSafe(mimeMessage, true);
                             transport(mimeMessage, recipients, transport, smtpConfig, mtaStatusInfo);
                             mailInterfaceMonitor.addUseTime(System.currentTimeMillis() - start);
                         } catch (final javax.mail.AuthenticationFailedException e) {
@@ -870,7 +940,7 @@ public final class SMTPTransport extends MailTransport implements MimeSupport {
                  */
                 smtpMessage.removeHeader("x-original-headers");
                 if (poisoned) {
-                    saveChangesSafe(smtpMessage);
+                    saveChangesSafe(smtpMessage, true);
                 } else {
                     final long start = System.currentTimeMillis();
                     final Transport transport = getSMTPSession().getTransport(SMTP);
@@ -879,7 +949,7 @@ public final class SMTPTransport extends MailTransport implements MimeSupport {
                         /*
                          * Save changes
                          */
-                        saveChangesSafe(smtpMessage);
+                        saveChangesSafe(smtpMessage, true);
                         /*
                          * TODO: Do encryption here
                          */
@@ -953,26 +1023,56 @@ public final class SMTPTransport extends MailTransport implements MimeSupport {
             transport.addTransportListener(new AddressAddingTransportListener(mtaInfo));
         }
 
+        // Grab listener chain instance
+        ListenerChain listenerChain = ListenerChain.getInstance();
+
         // Try to send the message
+        MimeMessage messageToSend = smtpMessage;
+        Exception exception = null;
         try {
-            transport.sendMessage(smtpMessage, recipients);
-            logMessageTransport(smtpMessage, smtpConfig);
+            // Check listener chain
+            Result result = listenerChain.onBeforeMessageTransport(messageToSend, session);
+            if (Reply.DENY == result.getReply()) {
+                throw MailExceptionCode.SEND_DENIED.create();
+            }
+            MimeMessage resultingMimeMessage = result.getMimeMessage();
+            if (null != resultingMimeMessage) {
+                messageToSend = resultingMimeMessage;
+            }
+
+            // Transport
+            transport.sendMessage(messageToSend, recipients);
+            logMessageTransport(messageToSend, smtpConfig);
+        } catch (OXException e) {
+            exception = e;
+            throw e;
         } catch (SMTPSendFailedException sendFailed) {
-            OXException oxe = MimeMailException.handleMessagingException(sendFailed, smtpConfig, session);
+            exception = sendFailed;
+            OXException oxe = MimeMailException.handleMessagingException(sendFailed, smtpConfig);
             if (null != mtaInfo) {
                 mtaInfo.setReturnCode(sendFailed.getReturnCode());
                 oxe.setArgument("mta_info", mtaInfo);
             }
+            Address[] validSentAddresses = sendFailed.getValidSentAddresses();
+            if (validSentAddresses != null && validSentAddresses.length > 0) {
+                try {
+                    oxe.setArgument("sent_message", MimeMessageConverter.convertMessage(messageToSend));
+                } catch (Exception e) {
+                    // Ignore
+                    LOG.debug("Failed to convert message after a message was partially sent", e);
+                }
+            }
             throw oxe;
         } catch (final MessagingException e) {
+            exception = e;
             if (e.getNextException() instanceof javax.activation.UnsupportedDataTypeException) {
                 // Check for "no object DCH for MIME type xxxxx/yyyy"
                 final String message = e.getNextException().getMessage();
-                if (toLowerCase(message).indexOf("no object dch") >= 0) {
+                if (message.toLowerCase().indexOf("no object dch") >= 0) {
                     // Not able to recover from JAF's "no object DCH for MIME type xxxxx/yyyy" error
                     // Perform the alternative transport with custom JAF DataHandler
                     LOG.warn(message.replaceFirst("[dD][cC][hH]", Matcher.quoteReplacement("javax.activation.DataContentHandler")));
-                    transportAlt(smtpMessage, recipients, transport, smtpConfig);
+                    transportAlt(messageToSend, recipients, transport, smtpConfig);
                     return;
                 }
             } else if (e.getNextException() instanceof IOException) {
@@ -980,7 +1080,12 @@ public final class SMTPTransport extends MailTransport implements MimeSupport {
                     throw MailExceptionCode.MAX_MESSAGE_SIZE_EXCEEDED.create(getSize(getMaxMailSize(), 2, false, true));
                 }
             }
-            throw MimeMailException.handleMessagingException(e, smtpConfig, session);
+            throw MimeMailException.handleMessagingException(e, smtpConfig);
+        } catch (RuntimeException e) {
+            exception = e;
+            throw MailExceptionCode.UNEXPECTED_ERROR.create(e, e.getMessage());
+        } finally {
+            listenerChain.onAfterMessageTransport(messageToSend, exception, session);
         }
     }
 
@@ -1222,7 +1327,7 @@ public final class SMTPTransport extends MailTransport implements MimeSupport {
         }
     }
 
-    private void saveChangesSafe(final MimeMessage mimeMessage) throws OXException {
+    private void saveChangesSafe(final MimeMessage mimeMessage, boolean keepMessageIdIfPresent) throws OXException {
         final HostnameService hostnameService = Services.getService(HostnameService.class);
         String hostName;
         if (null == hostnameService) {
@@ -1233,7 +1338,7 @@ public final class SMTPTransport extends MailTransport implements MimeSupport {
         if (null == hostName) {
             hostName = getHostName();
         }
-        MimeMessageConverter.saveChanges(mimeMessage, hostName);
+        MimeMessageUtility.saveChanges(mimeMessage, hostName, keepMessageIdIfPresent);
         // Check whether to remove MIME-Version headers from sub-parts
         if (TransportProperties.getInstance().isRemoveMimeVersionInSubParts()) {
             /*-
